# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of Montgomery and Shenzhen databases."""

import os

from ....data.datamodule import ConcatDataModule
from ..montgomery.datamodule import RawDataLoader as MontgomeryLoader
from ..montgomery.datamodule import make_split as make_montgomery_split
from ..shenzhen.datamodule import RawDataLoader as ShenzhenLoader
from ..shenzhen.datamodule import make_split as make_shenzhen_split


class DataModule(ConcatDataModule):
    """Aggregated DataModule composed of Montgomery and Shenzhen datasets.

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        montgomery_loader = MontgomeryLoader()
        montgomery_split = make_montgomery_split(split_filename)
        shenzhen_loader = ShenzhenLoader()
        shenzhen_split = make_shenzhen_split(split_filename)

        super().__init__(
            splits={
                "train": [
                    (montgomery_split["train"], montgomery_loader),
                    (shenzhen_split["train"], shenzhen_loader),
                ],
                "validation": [
                    (montgomery_split["validation"], montgomery_loader),
                    (shenzhen_split["validation"], shenzhen_loader),
                ],
                "test": [
                    (montgomery_split["test"], montgomery_loader),
                    (shenzhen_split["test"], shenzhen_loader),
                ],
            },
            database_name=__package__.split(".")[-1],
            split_name=os.path.splitext(split_filename)[0],
        )
