# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of Montgomery, Shenzhen, Indian, and TBX11k datasets."""

import os

from ....data.datamodule import ConcatDataModule
from ..indian.datamodule import CONFIGURATION_KEY_DATADIR as INDIAN_KEY_DATADIR
from ..indian.datamodule import RawDataLoader as IndianLoader
from ..indian.datamodule import make_split as make_indian_split
from ..montgomery.datamodule import RawDataLoader as MontgomeryLoader
from ..montgomery.datamodule import make_split as make_montgomery_split
from ..shenzhen.datamodule import RawDataLoader as ShenzhenLoader
from ..shenzhen.datamodule import make_split as make_shenzhen_split
from ..tbx11k.datamodule import RawDataLoader as TBX11kLoader
from ..tbx11k.datamodule import make_split as make_tbx11k_split


class DataModule(ConcatDataModule):
    """Aggregated DataModule composed of Montgomery, Shenzhen, Indian, and
    TBX11k datasets.

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    tbx11k_split_filename
        Name of the .json file from tbx11k containing the split to load.
    """

    def __init__(self, split_filename: str, tbx11k_split_filename: str):
        montgomery_loader = MontgomeryLoader()
        montgomery_split = make_montgomery_split(split_filename)
        shenzhen_loader = ShenzhenLoader()
        shenzhen_split = make_shenzhen_split(split_filename)
        indian_loader = IndianLoader(INDIAN_KEY_DATADIR)
        indian_split = make_indian_split(split_filename)
        tbx11k_loader = TBX11kLoader()
        tbx11k_split = make_tbx11k_split(tbx11k_split_filename)

        super().__init__(
            splits={
                "train": [
                    (montgomery_split["train"], montgomery_loader),
                    (shenzhen_split["train"], shenzhen_loader),
                    (indian_split["train"], indian_loader),
                    (tbx11k_split["train"], tbx11k_loader),
                ],
                "validation": [
                    (montgomery_split["validation"], montgomery_loader),
                    (shenzhen_split["validation"], shenzhen_loader),
                    (indian_split["validation"], indian_loader),
                    (tbx11k_split["validation"], tbx11k_loader),
                ],
                "test": [
                    (montgomery_split["test"], montgomery_loader),
                    (shenzhen_split["test"], shenzhen_loader),
                    (indian_split["test"], indian_loader),
                    (tbx11k_split["test"], tbx11k_loader),
                ],
            },
            database_name=__package__.split(".")[-1],
            split_name=(
                os.path.splitext(split_filename)[0]
                + "+"
                + os.path.splitext(tbx11k_split_filename)[0]
            ),
        )
