# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Aggregated DataModule composed of NIH-CXR-14 and PadChest databases."""

import os

from ....data.datamodule import ConcatDataModule
from ..nih_cxr14.datamodule import RawDataLoader as CXR14Loader
from ..nih_cxr14.datamodule import make_split as make_cxr14_split
from ..padchest.datamodule import RawDataLoader as PadchestLoader
from ..padchest.datamodule import make_split as make_padchest_split


class DataModule(ConcatDataModule):
    """Aggregated dataset composed of NIH CXR14 relabeld and PadChest
    (normalized) datasets.

    Parameters
    ----------
    cxr14_split_filename
        Name of the .json file from crx14 containing the split to load.
    padchest_split_filename
        Name of the .json file from padchest containing the split to load.
    """

    def __init__(self, cxr14_split_filename: str, padchest_split_filename):
        cxr14_loader = CXR14Loader()
        cxr14_split = make_cxr14_split(cxr14_split_filename)
        padchest_loader = PadchestLoader()
        padchest_split = make_padchest_split(padchest_split_filename)

        super().__init__(
            splits={
                "train": [
                    (cxr14_split["train"], cxr14_loader),
                    (padchest_split["train"], padchest_loader),
                ],
                "validation": [
                    (cxr14_split["validation"], cxr14_loader),
                    (padchest_split["validation"], padchest_loader),
                ],
                "test": [
                    (cxr14_split["test"], cxr14_loader),
                    # there is no test set on padchest
                    # (padchest_split["test"], padchest_loader),
                ],
            },
            database_name=__package__.split(".")[-1],
            split_name=(
                os.path.splitext(cxr14_split_filename)[0]
                + "+"
                + os.path.splitext(padchest_split_filename)[0]
            ),
        )
